#include <alloca.h>
#include <stdarg.h>
#include <stdlib.h>

#include "../include/uCTest/uCTest.h"

#include "System.h"
#include "messages.h"

static inline unsigned long __beginClkMeasure( unsigned long *clk ) {
    ( ( void ) clk );
    uctest_clkMeasureStart();
    return 1;
}

static inline void __endClkMeasure( unsigned long **__v ) {
    uctest_clkMeasureStop( *__v );
}

static inline unsigned long __beginClkMeasurePause( unsigned long *clk ) {
    uctest_clkMeasureStop( clk );
    return 1;
}

static inline void __endClkMeasurePause( unsigned long **__v ) {
    ( ( void ) __v );
    uctest_clkMeasureStart();
}

#ifdef CONFIC_COUNT_CYCLES
#define uctest_clockCycles( clocks )                                                           \
    for( unsigned long *__clkPtr                                                               \
         __attribute__( ( cleanup( __endClkMeasure ) ) ) = &clocks,                            \
                                                       __ToDo = __beginClkMeasure( __clkPtr ); \
         __ToDo; __ToDo = 0 )

#define uctest_clockCyclesPause( clocks )                                                     \
    for( unsigned long *__clkPtr __attribute__(                                               \
             ( cleanup( __endClkMeasurePause ) ) ) = &clocks,                                 \
                                                 __ToDo = __beginClkMeasurePause( __clkPtr ); \
         __ToDo; __ToDo = 0 )
#else
#define uctest_clockCycles( clocks ) \
    ( ( void ) clocks );             \
    clocks = 0;                      \
    for( int __ToDo = 1; __ToDo; __ToDo = 0 )

#define uctest_clockCyclesPause( clocks ) \
    ( ( void ) clocks );                  \
    clocks = 0;                           \
    for( int __ToDo = 1; __ToDo; __ToDo = 0 )
#endif

enum uctest_test_status {
    UCTEST_TEST_NOT_STARTED,
    UCTEST_TEST_SUCCESS,
    UCTEST_TEST_FAILURE,
    UCTEST_TEST_INVALID
};

struct uctest_test_rt {
    const __flash struct uctest_test *test;
    enum uctest_test_status status;
    unsigned long cycles;
};

static struct uctest_test **tests_list;
static int suites_count;
static int tests_count;

static struct uctest_test_rt *current_test_run;

void uctest_system_setup( void ) __attribute__( ( weak ) );

void uctest_system_teardown( void ) __attribute__( ( weak ) );

static int uctest_searchStack( const __flash char *name, void *ptr, int count ) {
    const __flash char **suitNames = ptr;
    for( int i = 0; i < count; i++ ) {
        if( name == *( suitNames + i ) ) {
            return 1;
        }
    }
    return 0;
}

static struct uctest_test_rt *
uctest_findInTests( struct uctest_test_rt *tests,
                    int ( *predicate )( struct uctest_test_rt *, void * ), void *data ) {
    struct uctest_test_rt *current = tests;
    for( int i = 0; i < tests_count; i++, current = ( tests + i ) ) {
        if( predicate( current, data ) ) {
            return current;
        }
    }
    return NULL;
}

static void uctest_collectRuntimeData( struct uctest_test_rt *tests, int *suitesCount,
                                       int *testsCount, int *testsSuccess ) {
    *suitesCount = 0;
    *testsCount = 0;
    void *stackPtr = uctest_getStackPtr();
    if( tests == NULL ) {
        const __flash struct uctest_test *current = *( tests_list );
        for( int i = 0; current; i++, current = *( tests_list - i ) ) {
            if( !uctest_searchStack( current->suite_name, uctest_getStackPtr() + 1,
                                     *suitesCount ) ) {
                uctest_stack_pushPtr( ( void * ) current->suite_name );
                ( *suitesCount )++;
            }
            if( current->tst_flags == UCTEST_TEST_SIMPLE ||
                current->tst_flags == UCTEST_TEST_FIXTURE ) {
                ( *testsCount )++;
            }
        }
    }
    else {
        struct uctest_test_rt *current = ( tests );
        for( int i = 0; i < tests_count; i++, current = ( tests + i ) ) {
            if( ( current->status == UCTEST_TEST_SUCCESS ||
                  current->status == UCTEST_TEST_FAILURE ) ) {
                ( *testsCount )++;
                if( current->status == UCTEST_TEST_SUCCESS && testsSuccess != NULL ) {
                    ( *testsSuccess )++;
                }
                if( !uctest_searchStack( current->test->suite_name, uctest_getStackPtr() + 1,
                                         *suitesCount ) ) {
                    uctest_stack_pushPtr( ( void * ) current->test->suite_name );
                    ( *suitesCount )++;
                }
            }
        }
    }
    uctest_setStackPtr( stackPtr );
}

void InitTests( void ) {
    tests_list = ( struct uctest_test ** ) ( uctest_stackStart - 1 );
    uctest_collectRuntimeData( NULL, &suites_count, &tests_count, NULL );
    uctest_printStartup( suites_count, tests_count );
    if( uctest_system_setup ) {
        uctest_printGlobalSetup();
        uctest_system_setup();
    }
}

static const __flash char *uctest_getPendingSuite( struct uctest_test_rt *tests ) {
    int predicate( struct uctest_test_rt * current, void *data ) {
        ( ( void ) data );
        return current->status == UCTEST_TEST_NOT_STARTED;
    }
    struct uctest_test_rt *res = uctest_findInTests( tests, predicate, NULL );
    return ( res != NULL ) ? res->test->suite_name : NULL;
}

static struct uctest_test_rt *uctest_getPendingTest( struct uctest_test_rt *tests,
                                                     const __flash char *suite ) {
    int predicate( struct uctest_test_rt * current, void *data ) {
        return ( ( ( const __flash char * ) data ) == current->test->suite_name &&
                 ( current->status == UCTEST_TEST_NOT_STARTED ) );
    }
    return uctest_findInTests( tests, predicate, ( void * ) suite );
}

static void uctest_getSetupTeardown( const __flash char *suite, void ( **setup )( void ),
                                     void ( **teardown )( void ) ) {
    *setup = NULL;
    *teardown = NULL;

    const __flash struct uctest_test *current = *( tests_list );
    for( int i = 0; current; i++, current = *( tests_list - i ) ) {
        if( current->suite_name != suite ) {
            continue;
        }
        if( current->tst_flags == UCTEST_TEST_SETUP ) {
            *setup = current->test_func;
        }
        if( current->tst_flags == UCTEST_TEST_TEARDOWN ) {
            *teardown = current->test_func;
        }
    }
}

static void uctest_runSuite( struct uctest_test_rt *tests, const __flash char *suite ) {
    int testsCount = 0;
    struct uctest_test_rt *current = ( tests );
    for( int i = 0; i < tests_count; i++, current = ( tests + i ) ) {
        if( suite == current->test->suite_name && current->status == UCTEST_TEST_NOT_STARTED ) {
            testsCount++;
        }
    }

    void ( *setup )( void );
    void ( *teardown )( void );
    uctest_getSetupTeardown( suite, &setup, &teardown );

    uctest_printSuiteStatsBegin( suite, testsCount );
    struct uctest_test_rt *current_test;
    while( ( current_test = uctest_getPendingTest( tests, suite ) ) != NULL ) {
        uctest_printTestBegin( current_test->test );
        if( current_test->test->tst_flags == UCTEST_TEST_FIXTURE ) {
            setup();
        }
        current_test_run = current_test;

        current_test->cycles = 0;
        uctest_clockCycles( current_test->cycles ) {
            current_test->test->test_func();
        }
        current_test_run = NULL;
        if( current_test->test->tst_flags == UCTEST_TEST_FIXTURE ) {
            teardown();
        }

        if( current_test->status != UCTEST_TEST_FAILURE ) {
            current_test->status = UCTEST_TEST_SUCCESS;
            uctest_printTestSuccess( current_test->test, current_test->cycles );
        }
        else {
            uctest_printTestFailure( current_test->test, current_test->cycles );
        }
    }

    testsCount = 0;
    unsigned long cyclesTotal = 0;
    current = ( tests );
    for( int i = 0; i < tests_count; i++, current = ( tests + i ) ) {
        if( suite == current->test->suite_name &&
            ( current->status == UCTEST_TEST_SUCCESS || current->status == UCTEST_TEST_FAILURE ) ) {
            testsCount++;
            cyclesTotal += current->cycles;
        }
    }

    uctest_printSuiteStatsEnd( suite, testsCount, cyclesTotal );
}

static void uctest_collectTests( struct uctest_test_rt *tests ) {
    const __flash struct uctest_test *current = *( tests_list );
    int k = 0;
    for( int i = 0; current; i++, current = *( tests_list - i ) ) {
        if( current->tst_flags == UCTEST_TEST_SIMPLE ||
            current->tst_flags == UCTEST_TEST_FIXTURE ) {
            ( tests + k )->test = current;
            ( tests + k )->status = UCTEST_TEST_NOT_STARTED;
            k++;
        }
    }
}

int run_all_tests( void ) {
    int testsRun = 0;
    int suitesRun = 0;
    int testsSuccess = 0;
    struct uctest_test_rt *tests = alloca( sizeof( struct uctest_test_rt ) * tests_count );
    uctest_collectTests( tests );
    const __flash char *current_suite;
    while( ( current_suite = uctest_getPendingSuite( tests ) ) != NULL ) {
        uctest_runSuite( tests, current_suite );
    }
    uctest_collectRuntimeData( tests, &suitesRun, &testsRun, &testsSuccess );

    if( uctest_system_teardown ) {
        uctest_printGlobalTeardown();
        uctest_system_teardown();
    }

    struct uctest_test_rt *current = tests;
    unsigned long totalCycles = 0;
    for( int i = 0; i < tests_count; i++, current = ( tests + i ) ) {
        if( current->status == UCTEST_TEST_FAILURE || current->status == UCTEST_TEST_SUCCESS ) {
            totalCycles += current->cycles;
        }
    }

    uctest_printRuntimeStatsEnd( suitesRun, testsRun, totalCycles );
    int testsFailure = testsRun - testsSuccess;
    uctest_printAllTestsPassed( testsSuccess );
    if( testsFailure ) {
        uctest_printFaildTestNum( testsFailure );
        current = tests;
        for( int i = 0; i < tests_count; i++, current = ( tests + i ) ) {
            if( current->status == UCTEST_TEST_FAILURE ) {
                uctest_printTestFailure( current->test, 0 );
            }
        }
        uctest_printFailedTestsAtEnd( testsFailure );
    }

    return testsFailure ? EXIT_FAILURE : EXIT_SUCCESS;
}

void __attribute__( ( format( printf, 3, 4 ) ) )
uctest_testFail( const __flash char *file, int line, const __flash char *fmt, ... ) {
    va_list valist;
    va_start( valist, fmt );
    if( current_test_run ) {
        uctest_clockCyclesPause( current_test_run->cycles ) {
            current_test_run->status = UCTEST_TEST_FAILURE;
            uctest_printFailureLocation( file, line );
            if( NULL != fmt ) {
                uctest_printUserMessage( fmt, valist );
            }
        }
    }
    va_end( valist );
}

static volatile void *ptr;
static volatile int abort_executed;

int uctest_assertDeath( void ( *func )( void ), const __flash char *msg ) {
    ( ( void ) msg ); // Ignore the message, in respect of memory we are not able to process these
    abort_executed = 0;
    static volatile stack_t *save_sp;
    save_sp =
        uctest_getStackPtr(); // Get the current stackpointer to restore it when the assert failed
    ptr = &&testAssert_abort; // This is the address to resume if the assert failed.
                              // It is not allowed to return from the abort.
    func();                   // Call the function to test the abort!

testAssert_abort:
    // restore the current frame
    uctest_setStackPtr( ( stack_t * ) save_sp );
    ptr = NULL;
    return abort_executed;
}

void abort( void ) {
    if( ptr ) {
        abort_executed = 1;
        goto *ptr;
    }
    else {
        exit( 134 );
    }
    while( 1 )
        ;
}

int uctest_main( void ) {
    InitTests();
    return RUN_ALL_TESTS();
}

int main( void ) __attribute__( ( weak, alias( "uctest_main" ) ) );
